# Copyright (C) 2024 qBraid
#
# This file is part of qbraid-qir
#
# Qbraid-qir is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for qbraid-qir, as per Section 15 of the GPL v3.

"""
This top level module contains the main qBraid QIR functionality.

.. currentmodule:: qbraid_qir

Functions
-----------

.. autosummary::
   :toctree: ../stubs/

   dumps


Exceptions
-----------

.. autosummary::
   :toctree: ../stubs/

   QbraidQirError
   QirConversionError

"""
import importlib
from typing import TYPE_CHECKING

from ._version import __version__
from .exceptions import QbraidQirError, QirConversionError
from .serialization import dumps

__all__ = [
    "__version__",
    "QbraidQirError",
    "QirConversionError",
    "dumps",
    "qasm3_to_qir",
    "cirq_to_qir",
]

_lazy = {"cirq": "cirq_to_qir", "qasm3": "qasm3_to_qir"}

if TYPE_CHECKING:
    from .cirq import cirq_to_qir
    from .qasm3 import qasm3_to_qir


def __getattr__(name):
    for mod_name, objects in _lazy.items():
        if name == mod_name:
            module = importlib.import_module(f".{mod_name}", __name__)
            globals()[mod_name] = module
            return module

        if name in objects:
            module = importlib.import_module(f".{mod_name}", __name__)
            obj = getattr(module, name)
            globals()[name] = obj
            return obj

    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")


def __dir__():
    return sorted(
        __all__ + list(_lazy.keys()) + [item for sublist in _lazy.values() for item in sublist]
    )



# Copyright (C) 2024 qBraid
#
# This file is part of qbraid-qir
#
# Qbraid-qir is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for qbraid-qir, as per Section 15 of the GPL v3.

"""
Module defining Cirq LLVM Module elements.
"""

import hashlib
from abc import ABCMeta, abstractmethod
from typing import Optional

import cirq
from pyqir import Context, Module


def generate_module_id(circuit: cirq.Circuit) -> str:
    """
    Generates a QIR module ID from a given Cirq circuit.

    This function serializes the Cirq circuit into a JSON string, computes its SHA-256 hash,
    and converts the hash into an alphanumeric string. The final name is a truncated version,
    prefixed with 'circuit-', to form a concise, semi-unique identifier.

    Args:
        circuit (cirq.Circuit): The Cirq circuit for which a unique name is to be generated.

    Returns:
        str: Alphanumeric module ID for the Cirq circuit

    """
    serialized_circuit = cirq.to_json(circuit)
    hash_object = hashlib.sha256(serialized_circuit.encode())
    hash_hex = hash_object.hexdigest()
    alphanumeric_hash = "".join(filter(str.isalnum, hash_hex))
    truncated_hash = alphanumeric_hash[:7]
    return f"circuit-{truncated_hash}"


class _CircuitElement(metaclass=ABCMeta):
    @abstractmethod
    def accept(self, visitor):
        pass


class _Register(_CircuitElement):
    def __init__(self, register: list[cirq.Qid]):
        self._register = register

    def accept(self, visitor):
        visitor.visit_register(self._register)


class _Operation(_CircuitElement):
    def __init__(self, operation: cirq.Operation):
        self._operation = operation

    def accept(self, visitor):
        visitor.visit_operation(self._operation)


class CirqModule:
    """
    A module representing a quantum circuit in Cirq using QIR.

    This class encapsulates a quantum circuit from Cirq and translates it into QIR format,
    maintaining information about quantum operations, qubits, and classical bits. It provides
    methods to interact with the underlying QIR module and circuit elements.

    Args:
        name (str): Name of the module.
        module (Module): QIR Module instance.
        num_qubits (int): Number of qubits in the circuit.
        elements (list[_CircuitElement]): list of circuit elements.

    Example:
        >>> circuit = cirq.Circuit()
        >>> cirq_module = CirqModule.from_circuit(circuit)
        >>> print(cirq_module.num_qubits)
    """

    def __init__(
        self,
        name: str,
        module: Module,
        num_qubits: int,
        elements: list[_CircuitElement],
    ):
        self._name = name
        self._module = module
        self._elements = elements
        self._num_qubits = num_qubits
        self._num_clbits = num_qubits  # create one classical bit for each qubit

    @property
    def name(self) -> str:
        """Returns the name of the module."""
        return self._name

    @property
    def module(self) -> Module:
        """Returns the QIR Module instance."""
        return self._module

    @property
    def num_qubits(self) -> int:
        """Returns the number of qubits in the circuit."""
        return self._num_qubits

    @property
    def num_clbits(self) -> int:
        """Returns the number of classical bits in the circuit."""
        return self._num_clbits

    @classmethod
    def from_circuit(cls, circuit: cirq.Circuit, module: Optional[Module] = None) -> "CirqModule":
        """Class method. Constructs a CirqModule from a given cirq.Circuit object
        and an optional QIR Module."""
        elements: list[_CircuitElement] = []

        # Register(s). Tentatively using cirq.Qid as input. Better approaches might exist tbd.
        elements.append(_Register(list(circuit.all_qubits())))

        # Operations
        for operation in circuit.all_operations():
            elements.append(_Operation(operation))

        if module is None:
            module = Module(Context(), generate_module_id(circuit))
        return cls(
            name="main",
            module=module,
            num_qubits=len(circuit.all_qubits()),
            elements=elements,
        )

    def accept(self, visitor):
        visitor.visit_cirq_module(self)
        for element in self._elements:
            element.accept(visitor)
        visitor.record_output(self)
        visitor.finalize()



# Copyright (C) 2024 qBraid
#
# This file is part of qbraid-qir
#
# Qbraid-qir is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for qbraid-qir, as per Section 15 of the GPL v3.

# pylint: disable=too-many-arguments

"""
Module defining Qasm3 Converter elements.

"""
from __future__ import annotations

import uuid
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import pyqasm.elements
    import pyqir


def generate_module_id() -> str:
    """
    Generates a QIR module ID from a given openqasm3 program.

    """
    # TODO: Consider a better approach of generating a unique identifier.
    generated_id = uuid.uuid1()
    return f"program-{generated_id}"


class QasmQIRModule:
    """
    A module representing an openqasm3 quantum program using QIR.

    Args:
        name (str): Name of the module.
        qasm_module (pyqasm.elements.Qasm3Module): The pyqasm qasm3 module.
        llvm_module (pyqir.Module): The QIR module.
    """

    def __init__(
        self,
        name: str,
        qasm_module: pyqasm.elements.Qasm3Module,
        llvm_module: pyqir.Module,
    ):
        self._name = name
        self._llvm_module = llvm_module
        self._qasm_program = qasm_module

    @property
    def name(self) -> str:
        """Returns the name of the module."""
        return self._name

    @property
    def llvm_module(self) -> pyqir.Module:
        """Returns the QIR Module instance."""
        return self._llvm_module

    @property
    def qasm_program(self) -> pyqasm.elements.Qasm3Module:
        """Returns the QASM3 program."""
        return self._qasm_program

    def accept(self, visitor):
        visitor.visit_qasm3_module(self)
        statements = self.qasm_program.unrolled_ast.statements
        for statement in statements:
            visitor.visit_statement(statement)
        visitor.record_output(self)
        visitor.finalize()



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# http://www.sphinx-doc.org/en/master/config

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

import qbraid_qir

# -- Project information -----------------------------------------------------

project = "qBraid"
copyright = "2024, qBraid Development Team"
author = "qBraid Development Team"

# The full version, including alpha/beta/rc tags
release = qbraid_qir.__version__

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.napoleon",
    "sphinx.ext.autodoc",
    "sphinx_autodoc_typehints",
    "sphinx.ext.autosummary",
    "sphinx.ext.todo",
    "sphinx.ext.mathjax",
    "sphinx.ext.coverage",
    "sphinx.ext.viewcode",
    "sphinx_copybutton",
]

# set_type_checking_flag = True
autodoc_member_order = "bysource"
autoclass_content = "both"
autodoc_mock_imports = ["cirq", "openqasm3", "pyqasm", "numpy", "numpy.typing"]
napoleon_numpy_docstring = False
todo_include_todos = True
mathjax_path = "https://cdn.jsdelivr.net/npm/mathjax@2/MathJax.js?config=TeX-AMS-MML_HTMLorMML"

# The master toctree document.
master_doc = "index"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = [
    "_build",
    "*.pytest_cache",
    "*.ipynb_checkpoints",
    "*tests",
]

# A boolean that decides whether module names are prepended to all object names
# (for object types where a “module” of some kind is defined), e.g. for
# py:function directives.
add_module_names = False

# A list of prefixes that are ignored for sorting the Python module index
# (e.g., if this is set to ['foo.'], then foo.bar is shown under B, not F).
# This can be handy if you document a project that consists of a single
# package. Works only for the HTML builder currently.
modindex_common_prefix = ["qbraid_qir."]

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = "sphinx_rtd_theme"

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
# html_theme_options = {
#     "collapse_navigation": False
# }

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

html_favicon = "_static/favicon.ico"
html_show_sphinx = False

html_css_files = ["css/s4defs-roles.css"]

# -- More customizations ----------------------------------------------------


# def skip_member(app, what, name, obj, skip, options):
#     print(app, what, name, obj, skip, options)
#     return True
#
#
# def setup(app):
#     app.connect('autodoc-skip-member', skip_member)



# Copyright (C) 2024 qBraid
#
# This file is part of qbraid-qir
#
# Qbraid-qir is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for qbraid-qir, as per Section 15 of the GPL v3.

"""
Module for processing Cirq circuits before conversion to QIR.

"""
import itertools
from typing import Iterable

import cirq

from .exceptions import CirqConversionError
from .opsets import map_cirq_op_to_pyqir_callable


def _decompose_gate_op(operation: cirq.Operation) -> Iterable[cirq.OP_TREE]:
    """Decomposes a single Cirq gate operation into a sequence of operations
    that are directly supported by PyQIR.

    Args:
        operation (cirq.Operation): The gate operation to decompose.

    Returns:
        Iterable[cirq.OP_TREE]: A list of decomposed gate operations.
    """
    try:
        # Try converting to PyQIR. If successful, keep the operation.
        _ = map_cirq_op_to_pyqir_callable(operation)
        return [operation]
    except CirqConversionError:
        pass
    new_ops = cirq.decompose_once(operation, flatten=True, default=[operation])
    if len(new_ops) == 1 and new_ops[0] == operation:
        raise CirqConversionError("Couldn't convert circuit to QIR gate set.")
    return list(itertools.chain.from_iterable(map(_decompose_gate_op, new_ops)))


def _decompose_unsupported_gates(circuit: cirq.Circuit) -> cirq.Circuit:
    """
    Decompose gates in a circuit that are not in the supported set.

    Args:
        circuit (cirq.Circuit): The quantum circuit to process.

    Returns:
        cirq.Circuit: A new circuit with unsupported gates decomposed.
    """
    new_circuit = cirq.Circuit()
    for moment in circuit:
        new_ops = []
        for operation in moment:
            if isinstance(operation, cirq.GateOperation):
                decomposed_ops = list(_decompose_gate_op(operation))
                new_ops.extend(decomposed_ops)
            elif isinstance(operation, cirq.ClassicallyControlledOperation):
                new_ops.append(operation)
            else:
                new_ops.append(operation)

        new_circuit.append(new_ops)
    return new_circuit


def preprocess_circuit(circuit: cirq.Circuit) -> cirq.Circuit:
    """
    Preprocesses a Cirq circuit to ensure that it is compatible with the QIR conversion.

    Args:
        circuit (cirq.Circuit): The Cirq circuit to preprocess.

    Returns:
        cirq.Circuit: The preprocessed Cirq circuit.

    """
    qubit_map = {qubit: cirq.LineQubit(i) for i, qubit in enumerate(circuit.all_qubits())}
    line_qubit_circuit = circuit.transform_qubits(lambda q: qubit_map[q])
    return _decompose_unsupported_gates(line_qubit_circuit)



# Copyright (C) 2024 qBraid
#
# This file is part of qbraid-qir
#
# Qbraid-qir is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for qbraid-qir, as per Section 15 of the GPL v3.

"""
Module defining exceptions for errors raised by qBraid QIR.

"""


class QbraidQirError(Exception):
    """Base class for errors raised by qbraid-qir."""


class QirConversionError(QbraidQirError):
    """Class for errors raised when converting quantum program to QIR."""



# Copyright (C) 2024 qBraid
#
# This file is part of qbraid-qir
#
# Qbraid-qir is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for qbraid-qir, as per Section 15 of the GPL v3.

"""
Module for exporting QIR to bitcode and LLVM IR files.

"""
from __future__ import annotations

import logging
import os
import pathlib
from typing import TYPE_CHECKING, Optional

from .exceptions import QbraidQirError

if TYPE_CHECKING:
    import pyqir


def dumps(module: pyqir.Module, output_dir: Optional[str] = None) -> None:
    """
    Saves the Quantum Intermediate Representation (QIR) of a given module
    to bitcode (.bc) and LLVM IR (.ll) files.

    Args:
        module: The module containing the QIR.
        output_dir (Optional[str]): The directory where the files will be saved.
            If not provided, defaults to the current working directory.

    Raises:
        QbraidQirError: If there's an error in writing the files.
    """
    filename_prefix = os.path.splitext(os.path.basename(module.source_filename))[0]

    file_dir = output_dir if output_dir else str(pathlib.Path.cwd())

    try:
        os.makedirs(file_dir, exist_ok=True)
    except OSError as err:
        raise QbraidQirError("Failed to create target directory") from err

    bc_file = os.path.join(file_dir, f"{filename_prefix}.bc")
    ll_file = os.path.join(file_dir, f"{filename_prefix}.ll")

    with open(bc_file, "wb") as file:  # pylint: disable=unspecified-encoding
        file.write(module.bitcode)
    logging.info("Saved to %s", bc_file)

    with open(ll_file, "w", encoding="utf-8") as file:
        file.write(str(module))
    logging.info("Saved to %s", ll_file)



# Copyright (C) 2024 qBraid
#
# This file is part of qbraid-qir
#
# Qbraid-qir is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for qbraid-qir, as per Section 15 of the GPL v3.

"""
Module containing Cirq QIR functionality.

.. currentmodule:: qbraid_qir.cirq

Functions
-----------

.. autosummary::
   :toctree: ../stubs/

   cirq_to_qir


Classes
---------

.. autosummary::
   :toctree: ../stubs/

   CirqModule
   BasicCirqVisitor

Exceptions
-----------

.. autosummary::
   :toctree: ../stubs/

   CirqConversionError

"""
from .convert import cirq_to_qir
from .elements import CirqModule
from .exceptions import CirqConversionError
from .visitor import BasicCirqVisitor

__all__ = ["cirq_to_qir", "CirqModule", "CirqConversionError", "BasicCirqVisitor"]



# Copyright (C) 2024 qBraid
#
# This file is part of qbraid-qir
#
# Qbraid-qir is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for qbraid-qir, as per Section 15 of the GPL v3.

"""
Module containing OpenQASM 3 QIR functionality.

.. currentmodule:: qbraid_qir.qasm3

Functions
-----------

.. autosummary::
   :toctree: ../stubs/

   qasm3_to_qir

Classes
---------

.. autosummary::
   :toctree: ../stubs/

   QasmQIRModule
   QasmQIRVisitor

Exceptions
-----------

.. autosummary::
   :toctree: ../stubs/

   Qasm3ConversionError

"""
from .convert import qasm3_to_qir
from .elements import QasmQIRModule
from .exceptions import Qasm3ConversionError
from .visitor import QasmQIRVisitor

__all__ = [
    "qasm3_to_qir",
    "QasmQIRModule",
    "Qasm3ConversionError",
    "QasmQIRVisitor",
]



# Copyright (C) 2024 qBraid
#
# This file is part of qbraid-qir
#
# Qbraid-qir is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for qbraid-qir, as per Section 15 of the GPL v3.

"""
Module defining exceptions for errors raised during QASM3 conversions.

"""
import logging
from typing import Optional, Type

from openqasm3.ast import Span

from qbraid_qir.exceptions import QirConversionError


class Qasm3ConversionError(QirConversionError):
    """Class for errors raised when converting an OpenQASM 3 program to QIR."""


def raise_qasm3_error(
    message: Optional[str] = None,
    err_type: Type[Exception] = Qasm3ConversionError,
    span: Optional[Span] = None,
    raised_from: Optional[Exception] = None,
) -> None:
    """Raises a QASM3 conversion error with optional chaining from another exception.

    Args:
        message: The error message. If not provided, a default message will be used.
        err_type: The type of error to raise.
        span: The span (location) in the QASM file where the error occurred.
        raised_from: Optional exception from which this error was raised (chaining).

    Raises:
        err_type: The error type initialized with the specified message and chained exception.
    """
    if span:
        logging.error(
            "Error at line %s, column %s in QASM file", span.start_line, span.start_column
        )

    if raised_from:
        raise err_type(message) from raised_from
    raise err_type(message)



# Copyright (C) 2024 qBraid
#
# This file is part of qbraid-qir
#
# Qbraid-qir is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for qbraid-qir, as per Section 15 of the GPL v3.

"""
Module mapping supported QASM gates/operations to pyqir functions.

"""
from typing import Callable, Union

import numpy as np
import pyqir
from pyqasm.linalg import kak_decomposition_angles

from .exceptions import Qasm3ConversionError


def id_gate(builder, qubits):
    pyqir._native.x(builder, qubits)
    pyqir._native.x(builder, qubits)


def u3_gate(
    builder,
    theta: Union[int, float],
    phi: Union[int, float],
    lam: Union[int, float],
    qubits,
):
    """
    Implements the U3 gate using the following decomposition:
         https://docs.quantum.ibm.com/api/qiskit/qiskit.circuit.library.UGate
         https://docs.quantum.ibm.com/api/qiskit/qiskit.circuit.library.PhaseGate

    Args:
        builder (pyqir._native.QirBuilder): The QIR builder.
        theta (Union[int, float]): The theta angle.
        phi (Union[int, float]): The phi angle.
        lam (Union[int, float]): The lambda angle.
        qubits: The qubits on which the gate is applied.

    Returns:
        None
    """
    pyqir._native.rz(builder, lam, qubits)
    pyqir._native.rx(builder, CONSTANTS_MAP["pi"] / 2, qubits)
    pyqir._native.rz(builder, theta + CONSTANTS_MAP["pi"], qubits)
    pyqir._native.rx(builder, CONSTANTS_MAP["pi"] / 2, qubits)
    pyqir._native.rz(builder, phi + CONSTANTS_MAP["pi"], qubits)
    # global phase - e^(i*(phi+lambda)/2) is missing in the above implementation


def u3_inv_gate(
    builder,
    theta: Union[int, float],
    phi: Union[int, float],
    lam: Union[int, float],
    qubits,
):
    """
    Implements the inverse of the U3 gate using the decomposition present in
    the u3_gate function.
    """
    pyqir._native.rz(builder, -1.0 * (phi + CONSTANTS_MAP["pi"]), qubits)
    pyqir._native.rx(builder, -1.0 * (CONSTANTS_MAP["pi"] / 2), qubits)
    pyqir._native.rz(builder, -1.0 * (theta + CONSTANTS_MAP["pi"]), qubits)
    pyqir._native.rx(builder, -1.0 * (CONSTANTS_MAP["pi"] / 2), qubits)
    pyqir._native.rz(builder, -1.0 * lam, qubits)


def u2_gate(builder, phi, lam, qubits):
    """
    Implements the U2 gate using the following decomposition:
        https://docs.quantum.ibm.com/api/qiskit/qiskit.circuit.library.U2Gate
    """
    u3_gate(builder, CONSTANTS_MAP["pi"] / 2, phi, lam, qubits)


def u2_inv_gate(builder, phi, lam, qubits):
    """
    Implements the inverse of the U2 gate using the decomposition present in
    the u2_gate function.
    """
    u3_inv_gate(builder, CONSTANTS_MAP["pi"] / 2, phi, lam, qubits)


def sx_gate(builder, qubits):
    """
    Implements the Sqrt(X) gate as a decomposition of other gates.
    """
    pyqir._native.rx(builder, CONSTANTS_MAP["pi"] / 2, qubits)


def sxdg_gate(builder, qubits):
    """
    Implements the conjugate transpose of the Sqrt(X) gate as a decomposition of other gates.
    """
    pyqir._native.rx(builder, -CONSTANTS_MAP["pi"] / 2, qubits)


def cv_gate(builder, qubit0, qubit1):
    """
    Implements the controlled V gate as a decomposition of other gates.
    """
    pyqir._native.x(builder, qubit0)
    pyqir._native.h(builder, qubit1)
    pyqir._native.cx(builder, qubit0, qubit1)
    pyqir._native.h(builder, qubit1)
    pyqir._native.rx(builder, CONSTANTS_MAP["pi"] / 4, qubit1)
    pyqir._native.h(builder, qubit1)
    pyqir._native.cx(builder, qubit0, qubit1)
    pyqir._native.t_adj(builder, qubit0)
    pyqir._native.h(builder, qubit1)
    pyqir._native.x(builder, qubit0)
    pyqir._native.rz(builder, -CONSTANTS_MAP["pi"] / 4, qubit1)


def cy_gate(builder, qubit0, qubit1):
    """
    Implements the CY gate as a decomposition of other gates.
    """
    pyqir._native.s_adj(builder, qubit1)
    pyqir._native.cx(builder, qubit0, qubit1)
    pyqir._native.s(builder, qubit1)


def xx_gate(builder, theta, qubit0, qubit1):
    """
    Implements the XX gate as a decomposition of other gates.
    """
    qubits = [qubit0, qubit1]
    pyqir._native.h(builder, qubits[0])
    pyqir._native.h(builder, qubits[1])
    pyqir._native.cz(builder, qubits[0], qubits[1])
    pyqir._native.h(builder, qubits[1])
    pyqir._native.rx(builder, theta, qubits[0])
    pyqir._native.h(builder, qubits[1])
    pyqir._native.cz(builder, qubits[0], qubits[1])
    pyqir._native.h(builder, qubits[0])
    pyqir._native.h(builder, qubits[1])


def xy_gate(builder, theta, qubit0, qubit1):
    """
    Implements the XY gate as a decomposition of other gates.
    """
    qubits = [qubit0, qubit1]
    pyqir._native.rx(builder, -theta / 2, qubits[0])
    pyqir._native.ry(builder, theta / 2, qubits[1])
    pyqir._native.ry(builder, theta / 2, qubits[0])
    pyqir._native.rx(builder, theta / 2, qubits[0])
    pyqir._native.cx(builder, qubits[1], qubits[0])
    pyqir._native.ry(builder, -theta / 2, qubits[0])
    pyqir._native.ry(builder, -theta / 2, qubits[1])
    pyqir._native.cx(builder, qubits[1], qubits[0])
    pyqir._native.rx(builder, theta / 2, qubits[0])
    pyqir._native.ry(builder, -theta / 2, qubits[1])
    pyqir._native.ry(builder, theta / 2, qubits[1])
    pyqir._native.rx(builder, -theta / 2, qubits[0])


def yy_gate(builder, theta, qubit0, qubit1):
    """
    Implements the YY gate as a decomposition of other gates.
    """
    qubits = [qubit0, qubit1]
    pyqir._native.rx(builder, theta / 2, qubits[0])
    pyqir._native.rx(builder, theta / 2, qubits[1])
    pyqir._native.cz(builder, qubits[0], qubits[1])
    pyqir._native.h(builder, qubits[1])
    pyqir._native.rx(builder, theta, qubits[1])
    pyqir._native.h(builder, qubits[1])
    pyqir._native.cz(builder, qubits[0], qubits[1])
    pyqir._native.rx(builder, -theta / 2, qubits[0])
    pyqir._native.rx(builder, -theta / 2, qubits[1])


def zz_gate(builder, theta, qubit0, qubit1):
    """
    Implements the ZZ gate as a decomposition of other gates.
    """
    qubits = [qubit0, qubit1]
    pyqir._native.cz(builder, qubits[0], qubits[1])
    pyqir._native.h(builder, qubits[1])
    pyqir._native.rz(builder, theta, qubits[1])
    pyqir._native.h(builder, qubits[1])
    pyqir._native.cz(builder, qubits[0], qubits[1])


def phaseshift_gate(builder, theta, qubit):
    """
    Implements the phase shift gate as a decomposition of other gates.
    """
    pyqir._native.h(builder, qubit)
    pyqir._native.rx(builder, theta, qubit)
    pyqir._native.h(builder, qubit)


def cswap_gate(builder, qubit0, qubit1, qubit2):
    """
    Implements the CSWAP gate as a decomposition of other gates.
    """
    qubits = [qubit0, qubit1, qubit2]
    pyqir._native.cx(builder, qubits[2], qubits[1])
    pyqir._native.h(builder, qubits[2])
    pyqir._native.cx(builder, qubits[1], qubits[2])
    pyqir._native.t_adj(builder, qubits[2])
    pyqir._native.cx(builder, qubits[0], qubits[2])
    pyqir._native.t(builder, qubits[2])
    pyqir._native.cx(builder, qubits[1], qubits[2])
    pyqir._native.t(builder, qubits[1])
    pyqir._native.t_adj(builder, qubits[2])
    pyqir._native.cx(builder, qubits[0], qubits[2])
    pyqir._native.cx(builder, qubits[0], qubits[1])
    pyqir._native.t(builder, qubits[2])
    pyqir._native.t(builder, qubits[0])
    pyqir._native.t_adj(builder, qubits[1])
    pyqir._native.h(builder, qubits[2])
    pyqir._native.cx(builder, qubits[0], qubits[1])
    pyqir._native.cx(builder, qubits[2], qubits[1])


def pswap_gate(builder, theta, qubit0, qubit1):
    """
    Implements the PSWAP gate as a decomposition of other gates.

    """
    qubits = [qubit0, qubit1]
    pyqir._native.swap(builder, qubits[0], qubits[1])
    pyqir._native.cx(builder, qubits[0], qubits[1])
    u3_gate(builder, 0, 0, theta, qubits[1])
    pyqir._native.cx(builder, qubits[0], qubits[1])


def cphaseshift_gate(builder, theta, qubit0, qubit1):
    """
    Implements the controlled phase shift gate as a decomposition of other gates.
    """
    qubits = [qubit0, qubit1]
    pyqir._native.h(builder, qubits[0])
    pyqir._native.rx(builder, theta / 2, qubits[0])
    pyqir._native.h(builder, qubits[0])
    pyqir._native.cx(builder, qubits[0], qubits[1])
    pyqir._native.h(builder, qubits[1])
    pyqir._native.rx(builder, -theta / 2, qubits[0])
    pyqir._native.h(builder, qubits[1])
    pyqir._native.cx(builder, qubits[0], qubits[1])
    pyqir._native.h(builder, qubits[1])
    pyqir._native.rx(builder, theta / 2, qubits[1])
    pyqir._native.h(builder, qubits[1])


def cphaseshift00_gate(builder, theta, qubit0, qubit1):
    """
    Implements the controlled phase shift 00 gate as a decomposition of other gates.

    """
    qubits = [qubit0, qubit1]
    pyqir._native.x(builder, qubits[0])
    pyqir._native.x(builder, qubits[1])
    u3_gate(builder, 0, 0, theta / 2, qubits[0])
    u3_gate(builder, 0, 0, theta / 2, qubits[1])
    pyqir._native.cx(builder, qubits[0], qubits[1])
    u3_gate(builder, 0, 0, -theta / 2, qubits[1])
    pyqir._native.cx(builder, qubits[0], qubits[1])
    pyqir._native.x(builder, qubits[0])
    pyqir._native.x(builder, qubits[1])


def cphaseshift01_gate(builder, theta, qubit0, qubit1):
    """
    Implements the controlled phase shift 01 gate as a decomposition of other gates.

    """
    qubits = [qubit0, qubit1]
    pyqir._native.x(builder, qubits[0])
    u3_gate(builder, 0, 0, theta / 2, qubits[1])
    u3_gate(builder, 0, 0, theta / 2, qubits[0])
    pyqir._native.cx(builder, qubits[0], qubits[1])
    u3_gate(builder, 0, 0, -theta / 2, qubits[1])
    pyqir._native.cx(builder, qubits[0], qubits[1])
    pyqir._native.x(builder, qubits[0])


def cphaseshift10_gate(builder, theta, qubit0, qubit1):
    """
    Implements the controlled phase shift 10 gate as a decomposition of other gates.

    """
    qubits = [qubit0, qubit1]
    u3_gate(builder, 0, 0, theta / 2, qubits[0])
    pyqir._native.x(builder, qubits[1])
    u3_gate(builder, 0, 0, theta / 2, qubits[1])
    pyqir._native.cx(builder, qubits[0], qubits[1])
    u3_gate(builder, 0, 0, -theta / 2, qubits[1])
    pyqir._native.cx(builder, qubits[0], qubits[1])
    pyqir._native.x(builder, qubits[1])


def gpi_gate(builder, phi, qubit):
    """
    Implements the gpi gate as a decomposition of other gates.
    """
    theta_0 = CONSTANTS_MAP["pi"]
    phi_0 = phi
    lambda_0 = -phi_0 + CONSTANTS_MAP["pi"]
    u3_gate(builder, theta_0, phi_0, lambda_0, qubit)


def gpi2_gate(builder, phi, qubit):
    """
    Implements the gpi2 gate as a decomposition of other gates.
    """
    theta_0 = CONSTANTS_MAP["pi"] / 2
    phi_0 = phi + 3 * CONSTANTS_MAP["pi"] / 2
    lambda_0 = -phi_0 + CONSTANTS_MAP["pi"] / 2
    u3_gate(builder, theta_0, phi_0, lambda_0, qubit)


# pylint: disable-next=too-many-arguments
def ms_gate(builder, phi0, phi1, theta, qubit0, qubit1):
    """
    Implements the Molmer Sorenson gate as a decomposition of other gates.
    """
    mat = np.array(
        [
            [
                np.cos(np.pi * theta),
                0,
                0,
                -1j * np.exp(-1j * 2 * np.pi * (phi0 + phi1)) * np.sin(np.pi * theta),
            ],
            [
                0,
                np.cos(np.pi * theta),
                -1j * np.exp(-1j * 2 * np.pi * (phi0 - phi1)) * np.sin(np.pi * theta),
                0,
            ],
            [
                0,
                -1j * np.exp(1j * 2 * np.pi * (phi0 - phi1)) * np.sin(np.pi * theta),
                np.cos(np.pi * theta),
                0,
            ],
            [
                -1j * np.exp(1j * 2 * np.pi * (phi0 + phi1)) * np.sin(np.pi * theta),
                0,
                0,
                np.cos(np.pi * theta),
            ],
        ]
    )
    angles = kak_decomposition_angles(mat)
    qubits = [qubit0, qubit1]

    u3_gate(builder, angles[0][0], angles[0][1], angles[0][2], qubits[0])
    u3_gate(builder, angles[1][0], angles[1][1], angles[1][2], qubits[1])
    sx_gate(builder, qubits[0])
    pyqir._native.cx(builder, qubits[0], qubits[1])
    pyqir._native.rx(builder, ((1 / 2) - 2 * theta) * CONSTANTS_MAP["pi"], qubits[0])
    pyqir._native.rx(builder, CONSTANTS_MAP["pi"] / 2, qubits[1])
    pyqir._native.cx(builder, qubits[1], qubits[0])
    sxdg_gate(builder, qubits[1])
    pyqir._native.s(builder, qubits[1])
    pyqir._native.cx(builder, qubits[0], qubits[1])
    u3_gate(builder, angles[2][0], angles[2][1], angles[2][2], qubits[0])
    u3_gate(builder, angles[3][0], angles[3][1], angles[3][2], qubits[1])


def ecr_gate(builder, qubit0, qubit1):
    """
    Implements the ECR gate as a decomposition of other gates.

    """
    qubits = [qubit0, qubit1]
    pyqir._native.s(builder, qubits[0])
    pyqir._native.rx(builder, CONSTANTS_MAP["pi"] / 2, qubits[1])
    pyqir._native.cx(builder, qubits[0], qubits[1])
    pyqir._native.x(builder, qubits[0])


def prx_gate(builder, theta, phi, qubit):
    """
    Implements the PRX gate as a decomposition of other gates.
    """
    theta_0 = theta
    phi_0 = CONSTANTS_MAP["pi"] / 2 - phi
    lambda_0 = -phi_0
    u3_gate(builder, theta_0, phi_0, lambda_0, qubit)


PYQIR_ONE_QUBIT_OP_MAP = {
    "id": id_gate,
    "h": pyqir._native.h,
    "x": pyqir._native.x,
    "y": pyqir._native.y,
    "z": pyqir._native.z,
    "s": pyqir._native.s,
    "t": pyqir._native.t,
    "sdg": pyqir._native.s_adj,
    "si": pyqir._native.s_adj,
    "tdg": pyqir._native.t_adj,
    "ti": pyqir._native.t_adj,
    "v": sx_gate,
    "sx": sx_gate,
    "vi": sxdg_gate,
    "sxdg": sxdg_gate,
}

PYQIR_ONE_QUBIT_ROTATION_MAP = {
    "rx": pyqir._native.rx,
    "ry": pyqir._native.ry,
    "rz": pyqir._native.rz,
    "u": u3_gate,
    "U": u3_gate,
    "u3": u3_gate,
    "U3": u3_gate,
    "U2": u2_gate,
    "u2": u2_gate,
    "prx": prx_gate,
    "phaseshift": phaseshift_gate,
    "p": phaseshift_gate,
    "gpi": gpi_gate,
    "gpi2": gpi2_gate,
}

PYQIR_TWO_QUBIT_OP_MAP = {
    "cx": pyqir._native.cx,
    "CX": pyqir._native.cx,
    "cnot": pyqir._native.cx,
    "cz": pyqir._native.cz,
    "swap": pyqir._native.swap,
    "cv": cv_gate,
    "cy": cy_gate,
    "xx": xx_gate,
    "xy": xy_gate,
    "yy": yy_gate,
    "zz": zz_gate,
    "pswap": pswap_gate,
    "cp": cphaseshift_gate,
    "cphaseshift": cphaseshift_gate,
    "cp00": cphaseshift00_gate,
    "cphaseshift00": cphaseshift00_gate,
    "cp01": cphaseshift01_gate,
    "cphaseshift01": cphaseshift01_gate,
    "cp10": cphaseshift10_gate,
    "cphaseshift10": cphaseshift10_gate,
    "ecr": ecr_gate,
    "ms": ms_gate,
}

PYQIR_THREE_QUBIT_OP_MAP = {
    "ccx": pyqir._native.ccx,
    "ccnot": pyqir._native.ccx,
    "cswap": cswap_gate,
}


def map_qasm_op_to_pyqir_callable(op_name: str) -> tuple[Callable, int]:
    """
    Map a QASM operation to a PyQIR callable.

    Args:
        op_name (str): The QASM operation name.

    Returns:
        tuple: A tuple containing the PyQIR callable and the number of qubits the operation acts on.

    Raises:
        Qasm3ConversionError: If the QASM operation is unsupported or undeclared.
    """
    qasm_op_mappings: list[tuple[dict, int]] = [
        (PYQIR_ONE_QUBIT_OP_MAP, 1),
        (PYQIR_ONE_QUBIT_ROTATION_MAP, 1),
        (PYQIR_TWO_QUBIT_OP_MAP, 2),
        (PYQIR_THREE_QUBIT_OP_MAP, 3),
    ]

    for mapping, qubits in qasm_op_mappings:
        if op_name in mapping:
            return mapping[op_name], qubits

    raise Qasm3ConversionError(f"Unsupported / undeclared QASM operation: {op_name}")


CONSTANTS_MAP = {
    "π": 3.141592653589793,
    "pi": 3.141592653589793,
    "ℇ": 2.718281828459045,
    "euler": 2.718281828459045,
    "τ": 6.283185307179586,
    "tau": 6.283185307179586,
}



# Copyright (C) 2024 qBraid
#
# This file is part of qbraid-qir
#
# Qbraid-qir is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for qbraid-qir, as per Section 15 of the GPL v3.

"""
Module containing Cirq to qBraid QIR conversion functions

"""
from typing import Optional

import cirq
from pyqir import Context, Module, qir_module

from .elements import CirqModule, generate_module_id
from .exceptions import CirqConversionError
from .passes import preprocess_circuit
from .visitor import BasicCirqVisitor


def cirq_to_qir(circuit: cirq.Circuit, name: Optional[str] = None, **kwargs) -> Module:
    """
    Converts a Cirq circuit to a PyQIR module.

    Args:
        circuit (cirq.Circuit): The Cirq circuit to convert.
        name (str, optional): Identifier for created QIR module. Auto-generated if not provided.

    Keyword Args:
        initialize_runtime (bool): Whether to perform quantum runtime environment initialization,
                                   default `True`.
        record_output (bool): Whether to record output calls for registers, default `True`

    Returns:
        The QIR ``pyqir.Module`` representation of the input Cirq circuit.

    Raises:
        TypeError: If the input is not a valid Cirq circuit.
        ValueError: If the input circuit is empty.
        CirqConversionError: If the conversion fails.
    """
    if not isinstance(circuit, cirq.Circuit):
        raise TypeError("Input quantum program must be of type cirq.Circuit.")

    if len(circuit) == 0:
        raise ValueError("Input quantum circuit must consist of at least one operation.")

    if name is None:
        name = generate_module_id(circuit)

    try:
        circuit = preprocess_circuit(circuit)
    except Exception as err:  # pylint: disable=broad-exception-caught
        raise CirqConversionError("Failed to preprocess circuit.") from err

    llvm_module = qir_module(Context(), name)
    module = CirqModule.from_circuit(circuit, llvm_module)

    visitor = BasicCirqVisitor(**kwargs)
    module.accept(visitor)

    error = llvm_module.verify()
    if error is not None:
        raise CirqConversionError(error)
    return llvm_module



# Copyright (C) 2024 qBraid
#
# This file is part of qbraid-qir
#
# Qbraid-qir is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for qbraid-qir, as per Section 15 of the GPL v3.

"""
Module containing OpenQASM to QIR conversion functions

"""
from typing import Optional, Union

import openqasm3
import pyqasm
from pyqir import Context, Module, qir_module

from .elements import QasmQIRModule, generate_module_id
from .exceptions import Qasm3ConversionError
from .visitor import QasmQIRVisitor


def qasm3_to_qir(
    program: Union[openqasm3.ast.Program, str],
    name: Optional[str] = None,
    external_gates: Optional[list[str]] = None,
    **kwargs,
) -> Module:
    """Converts an OpenQASM 3 program to a PyQIR module.

    Args:
        program (openqasm3.ast.Program or str): The OpenQASM 3 program to convert.
        name (str, optional): Identifier for created QIR module. Auto-generated if not provided.
        external_gates (list[str], optional): A list of custom gate names that are not natively
            recognized by pyqasm but should be treated as valid during program unrolling.

    Keyword Args:
        initialize_runtime (bool): Whether to perform quantum runtime environment initialization,
            defaults to `True`.
        record_output (bool): Whether to record output calls for registers, defaults to `True`.

    Returns:
        The QIR ``pyqir.Module`` representation of the input OpenQASM 3 program.

    Raises:
        TypeError: If the input is not a valid OpenQASM 3 program.
        Qasm3ConversionError: If the conversion fails.
    """
    if isinstance(program, openqasm3.ast.Program):
        program = openqasm3.dumps(program)

    elif not isinstance(program, str):
        raise TypeError("Input quantum program must be of type openqasm3.ast.Program or str.")

    qasm3_module = pyqasm.loads(program)
    qasm3_module.unroll(external_gates=external_gates)
    if name is None:
        name = generate_module_id()
    llvm_module = qir_module(Context(), name)

    final_module = QasmQIRModule(name, qasm3_module, llvm_module)

    visitor = QasmQIRVisitor(external_gates=external_gates, **kwargs)
    final_module.accept(visitor)

    err = llvm_module.verify()
    if err is not None:
        raise Qasm3ConversionError(err)
    return llvm_module



# Copyright (C) 2024 qBraid
#
# This file is part of qbraid-qir
#
# Qbraid-qir is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for qbraid-qir, as per Section 15 of the GPL v3.

"""
Module defining CirqVisitor.

"""
import logging
from abc import ABCMeta, abstractmethod

import cirq
import pyqir
import pyqir._native
import pyqir.rt
from pyqir import BasicBlock, Builder, Constant, IntType, PointerType

from .elements import CirqModule
from .opsets import map_cirq_op_to_pyqir_callable

logger = logging.getLogger(__name__)


class CircuitElementVisitor(metaclass=ABCMeta):
    @abstractmethod
    def visit_register(self, qids):
        pass

    @abstractmethod
    def visit_operation(self, operation):
        pass


class BasicCirqVisitor(CircuitElementVisitor):
    """A visitor for basic cirq.Circuit elements.

    This class is designed to traverse and interact with elements in a quantum circuit.

    Args:
        initialize_runtime (bool): If True, quantum runtime will be initialized. Defaults to True.
        record_output (bool): If True, output of the circuit will be recorded. Defaults to True.
    """

    def __init__(self, initialize_runtime: bool = True, record_output: bool = True):
        self._module: pyqir.Module
        self._builder: pyqir.Builder
        self._entry_point: str
        self._qubit_labels: dict[cirq.Qid, int] = {}
        self._measured_qubits: dict = {}
        self._initialize_runtime = initialize_runtime
        self._record_output = record_output

    def visit_cirq_module(self, module: CirqModule) -> None:
        logger.debug("Visiting Cirq module '%s' (%d)", module.name, module.num_qubits)
        self._module = module.module
        context = self._module.context
        entry = pyqir.entry_point(self._module, module.name, module.num_qubits, module.num_clbits)

        self._entry_point = entry.name
        self._builder = Builder(context)
        self._builder.insert_at_end(BasicBlock(context, "entry", entry))

        if self._initialize_runtime is True:
            i8p = PointerType(IntType(context, 8))
            nullptr = Constant.null(i8p)
            pyqir.rt.initialize(self._builder, nullptr)

    @property
    def entry_point(self) -> str:
        return self._entry_point

    def finalize(self) -> None:
        self._builder.ret(None)

    def record_output(self, module: CirqModule) -> None:
        if self._record_output is False:
            return

        i8p = PointerType(IntType(self._module.context, 8))

        for i in range(module.num_qubits):
            result_ref = pyqir.result(self._module.context, i)
            pyqir.rt.result_record_output(self._builder, result_ref, Constant.null(i8p))

    def visit_register(self, qids: list[cirq.Qid]) -> None:
        logger.debug("Visiting qids '%s'", str(qids))

        if not all(isinstance(x, cirq.Qid) for x in qids):
            raise TypeError("All elements in the list must be of type cirq.Qid.")

        self._qubit_labels.update({bit: n + len(self._qubit_labels) for n, bit in enumerate(qids)})
        logger.debug("Added labels for qubits %s", str(qids))

    def visit_operation(self, operation: cirq.Operation) -> None:
        qlabels = [self._qubit_labels[bit] for bit in operation.qubits]

        qubits = [pyqir.qubit(self._module.context, n) for n in qlabels]
        results = [pyqir.result(self._module.context, n) for n in qlabels]

        def handle_measurement(pyqir_func):
            logger.debug("Visiting measurement operation '%s'", str(operation))
            for qubit, result in zip(qubits, results):
                self._measured_qubits[pyqir.qubit_id(qubit)] = True
                pyqir_func(self._builder, qubit, result)

        # dealing with conditional gates
        if isinstance(operation, cirq.ClassicallyControlledOperation):
            op_conds = operation._conditions  # list of measurement keys
            conditions = [
                pyqir.result(self._module.context, int(op_conds[i].keys[0].name))
                for i in range(len(op_conds))
            ]
            regular_op = operation.without_classical_controls()
            temp_pyqir_func, op_str = map_cirq_op_to_pyqir_callable(regular_op)

            # pylint: disable=unnecessary-lambda-assignment
            if op_str in ["Rx", "Ry", "Rz"]:
                pyqir_func = lambda: temp_pyqir_func(
                    self._builder,
                    operation._sub_operation.gate._rads,  # type: ignore[union-attr]
                    *qubits,
                )
            else:
                pyqir_func = lambda: temp_pyqir_func(self._builder, *qubits)

            def _branch(conds, pyqir_func):
                if len(conds) == 0:
                    temp_id, _ = map_cirq_op_to_pyqir_callable(cirq.I)
                    passable_identity = lambda: temp_id(self._builder, *qubits)
                    return passable_identity
                return pyqir._native.if_result(
                    self._builder,
                    conds[0],
                    zero=_branch(conds[1:], pyqir_func),
                    one=pyqir_func,
                )

            _branch(conditions, pyqir_func)
        else:
            pyqir_func, op_str = map_cirq_op_to_pyqir_callable(operation)

            if op_str.startswith("measure"):
                handle_measurement(pyqir_func)
            elif op_str in ["Rx", "Ry", "Rz"]:
                pyqir_func(self._builder, operation.gate._rads, *qubits)  # type: ignore[union-attr]
            else:
                pyqir_func(self._builder, *qubits)

    def ir(self) -> str:
        return str(self._module)



# Copyright (C) 2024 qBraid
#
# This file is part of qbraid-qir
#
# Qbraid-qir is free software released under the GNU General Public License v3
# or later. You can redistribute and/or modify it under the terms of the GPL v3.
# See the LICENSE file in the project root or <https://www.gnu.org/licenses/gpl-3.0.html>.
#
# THERE IS NO WARRANTY for qbraid-qir, as per Section 15 of the GPL v3.

# pylint: disable=too-many-instance-attributes,too-many-lines,too-many-branches

"""
Module defining Qasm3 Visitor.

"""
import logging
from typing import Any, Union

import openqasm3.ast as qasm3_ast
import pyqir
import pyqir._native
import pyqir.rt
from openqasm3.ast import UnaryOperator

from .elements import QasmQIRModule
from .exceptions import raise_qasm3_error
from .maps import map_qasm_op_to_pyqir_callable

logger = logging.getLogger(__name__)


class QasmQIRVisitor:
    """A visitor for converting OpenQASM 3 programs to QIR.

    This class is designed to traverse and interact with statements in an OpenQASM program.

    Args:
        initialize_runtime (bool): If True, quantum runtime will be initialized. Defaults to True.
        record_output (bool): If True, output of the circuit will be recorded. Defaults to True.
        external_gates (list[str]): List of custom gates that should not be unrolled.
                                    Instead, these gates are marked for external linkage, as
                                    qir-functions with the name "__quantum__qis__<GateName>__body"
    """

    def __init__(
        self,
        initialize_runtime: bool = True,
        record_output: bool = True,
        external_gates: list[str] | None = None,
    ):
        self._llvm_module: pyqir.Module
        self._builder: pyqir.Builder
        self._entry_point: str = ""
        self._qubit_labels: dict[str, int] = {}
        self._clbit_labels: dict[str, int] = {}
        self._global_qreg_size_map: dict[str, int] = {}
        self._global_creg_size_map: dict[str, int] = {}
        self._custom_gates: dict[str, qasm3_ast.QuantumGateDefinition] = {}
        self._barrier_qubits: set[pyqir.Constant] = set()
        self._initialize_runtime: bool = initialize_runtime
        self._record_output: bool = record_output

        if external_gates is None:
            external_gates = []
        self._external_gates_map: dict[str, pyqir.Function | None] = {
            external_gate: None for external_gate in external_gates
        }

    def visit_qasm3_module(self, module: QasmQIRModule) -> None:
        """
        Visit a Qasm3 module.

        Args:
            module (Qasm3Module): The module to visit.

        Returns:
            None
        """
        qasm3_module = module.qasm_program
        logger.debug("Visiting Qasm3 module '%s' (%d)", module.name, qasm3_module.num_qubits)
        self._llvm_module = module.llvm_module
        context = self._llvm_module.context
        entry = pyqir.entry_point(
            self._llvm_module, module.name, qasm3_module.num_qubits, qasm3_module.num_clbits
        )

        self._entry_point = entry.name
        self._builder = pyqir.Builder(context)
        self._builder.insert_at_end(pyqir.BasicBlock(context, "entry", entry))

        if self._initialize_runtime is True:
            i8p = pyqir.PointerType(pyqir.IntType(context, 8))
            nullptr = pyqir.Constant.null(i8p)
            pyqir.rt.initialize(self._builder, nullptr)

    @property
    def entry_point(self) -> str:
        return self._entry_point

    def finalize(self) -> None:
        self._check_and_apply_barrier()  # to check if we have an incomplete barrier at program end
        self._builder.ret(None)

    def record_output(self, module: QasmQIRModule) -> None:
        if self._record_output is False:
            return
        i8p = pyqir.PointerType(pyqir.IntType(self._llvm_module.context, 8))
        for i in range(module.qasm_program.num_qubits):
            result_ref = pyqir.result(self._llvm_module.context, i)
            pyqir.rt.result_record_output(self._builder, result_ref, pyqir.Constant.null(i8p))

    def _visit_register(
        self, register: Union[qasm3_ast.QubitDeclaration, qasm3_ast.ClassicalDeclaration]
    ) -> None:
        """Visit a register statement.

        Args:
            register (QubitDeclaration|ClassicalDeclaration): The register name and size.

        Returns:
            None
        """
        logger.debug("Visiting register '%s'", str(register))
        is_qubit = isinstance(register, qasm3_ast.QubitDeclaration)

        current_size = len(self._qubit_labels) if is_qubit else len(self._clbit_labels)
        if is_qubit:
            register_size = (
                1 if register.size is None else register.size.value  # type: ignore[union-attr]
            )
        else:
            register_size = (
                1
                if register.type.size is None  # type: ignore[union-attr]
                else register.type.size.value  # type: ignore[union-attr]
            )
        register_name = (
            register.qubit.name  # type: ignore[union-attr]
            if is_qubit
            else register.identifier.name  # type: ignore[union-attr]
        )

        size_map = self._global_qreg_size_map if is_qubit else self._global_creg_size_map
        label_map = self._qubit_labels if is_qubit else self._clbit_labels

        for i in range(register_size):
            size_map[f"{register_name}"] = register_size
            label_map[f"{register_name}_{i}"] = current_size + i

        logger.debug("Added labels for register '%s'", str(register))

    def _get_op_bits(self, operation: Any, qubits: bool = True) -> list[pyqir.Constant]:
        """Get the quantum / classical bits for the operation.

        Args:
            operation (Any): The operation to get qubits for.
            reg_size_map (dict): The size map of the registers in scope.
            qubits (bool): Whether the bits are quantum bits or classical bits. Defaults to True.

        Returns:
            Unionlist[pyqir.Constant] : The bits for the operation.
        """
        qir_bits = []
        bit_list = []
        if isinstance(operation, qasm3_ast.QuantumMeasurementStatement):
            assert operation.target is not None
            bit_list = [operation.measure.qubit] if qubits else [operation.target]
        else:
            bit_list = (
                operation.qubits if isinstance(operation.qubits, list) else [operation.qubits]
            )

        for bit in bit_list:
            # as we have unrolled qasm3, we can assume that the bit is an IndexedIdentifier
            assert isinstance(bit, qasm3_ast.IndexedIdentifier)
            reg_name = bit.name.name

            assert isinstance(bit.indices, list) and len(bit.indices) == 1
            assert isinstance(bit.indices[0], list) and len(bit.indices[0]) == 1
            assert isinstance(bit.indices[0][0], qasm3_ast.IntegerLiteral)
            bit_id = bit.indices[0][0].value
            bit_ids = [bit_id]

            label_map = self._qubit_labels if qubits else self._clbit_labels
            reg_ids = [label_map[f"{reg_name}_{bit_id}"] for bit_id in bit_ids]

            qir_bits.extend(
                [
                    (
                        pyqir.qubit(self._llvm_module.context, bit_id)
                        if qubits
                        else pyqir.result(self._llvm_module.context, bit_id)
                    )
                    for bit_id in reg_ids
                ]
            )

        return qir_bits

    def _visit_measurement(self, statement: qasm3_ast.QuantumMeasurementStatement) -> None:
        """Visit a measurement statement element.

        Args:
            statement (qasm3_ast.QuantumMeasurementStatement): The measurement statement to visit.

        Returns:
            None
        """
        logger.debug("Visiting measurement statement '%s'", str(statement))

        source = statement.measure.qubit
        target = statement.target
        assert source and target
        source_ids = self._get_op_bits(statement, qubits=True)
        target_ids = self._get_op_bits(statement, qubits=False)
        for src_id, tgt_id in zip(source_ids, target_ids):
            pyqir._native.mz(self._builder, src_id, tgt_id)  # type: ignore[arg-type]

    def _visit_reset(self, statement: qasm3_ast.QuantumReset) -> None:
        """Visit a reset statement element.

        Args:
            statement (qasm3_ast.QuantumReset): The reset statement to visit.

        Returns:
            None
        """
        logger.debug("Visiting reset statement '%s'", str(statement))
        qubit_ids = self._get_op_bits(statement, True)

        for qid in qubit_ids:
            # qid is of type Constant which is inherited from Value, so we ignore the type error
            pyqir._native.reset(self._builder, qid)  # type: ignore[arg-type]

    def _barrier_applicable(self) -> bool:
        """Check if the barrier operation is applicable.

        Args:
            None

        Returns:
            bool: Whether the barrier operation is applicable.
        """
        total_qubit_count = sum(self._global_qreg_size_map.values())
        return len(self._barrier_qubits) == total_qubit_count

    def _check_and_apply_barrier(self) -> None:
        """Apply the barrier operation.

        Returns:
            None
        """
        if len(self._barrier_qubits) == 0:
            return

        if self._barrier_applicable():
            pyqir._native.barrier(self._builder)
            self._barrier_qubits.clear()
        else:
            raise_qasm3_error(
                "Barrier operation on a qubit subset is not supported in pyqir",
                err_type=NotImplementedError,
            )

    # pylint: disable=unused-argument
    def _visit_barrier(self, barrier: qasm3_ast.QuantumBarrier) -> None:
        """Visit a barrier statement element.

        Args:
            statement (qasm3_ast.QuantumBarrier): The barrier statement to visit.
        Returns:
            None
        """
        barrier_qubit = self._get_op_bits(barrier, qubits=True)
        self._barrier_qubits.update(barrier_qubit)

        # try to apply barrier in case all qubits are covered here itself
        if self._barrier_applicable():
            pyqir._native.barrier(self._builder)
            self._barrier_qubits.clear()

    def _get_op_parameters(self, operation: qasm3_ast.QuantumGate) -> list[float]:
        """Get the parameters for the operation.

        Args:
            operation (qasm3_ast.QuantumGate): The operation to get parameters for.

        Returns:
            list[float]: The parameters for the operation.
        """
        param_list = []
        for param in operation.arguments:
            assert hasattr(param, "value")
            param_value = param.value
            param_list.append(param_value)

        return param_list

    def _visit_basic_gate_operation(self, operation: qasm3_ast.QuantumGate) -> None:
        """Visit a gate operation element.

        Args:
            operation (qasm3_ast.QuantumGate): The gate operation to visit.


        Returns:
            None

        Raises:
            Qasm3ConversionError: If the number of qubits is invalid.

        """

        logger.debug("Visiting basic gate operation '%s'", str(operation))
        op_name: str = operation.name.name
        op_qubits = self._get_op_bits(operation)
        qir_func, op_qubit_count = map_qasm_op_to_pyqir_callable(op_name)
        op_parameters = None

        if len(operation.arguments) > 0:  # parametric gate
            op_parameters = self._get_op_parameters(operation)

        for i in range(0, len(op_qubits), op_qubit_count):
            # we apply the gate on the qubit subset linearly
            qubit_subset = op_qubits[i : i + op_qubit_count]
            if op_parameters is not None:
                qir_func(self._builder, *op_parameters, *qubit_subset)
            else:
                qir_func(self._builder, *qubit_subset)

    def _visit_external_gate_operation(self, operation: qasm3_ast.QuantumGate) -> None:
        """Visit an external gate operation element.

        Args:
            operation (qasm3_ast.QuantumGate): The gate operation to visit.


        Returns:
            None

        Raises:
            Qasm3ConversionError: If the number of qubits is invalid.

        """
        logger.debug("Visiting external gate operation '%s'", str(operation))
        op_name: str = operation.name.name
        op_qubits = self._get_op_bits(operation)
        op_qubit_count = len(op_qubits)

        if len(operation.modifiers) > 0:
            raise_qasm3_error(
                "Modifiers on externally linked gates are not supported in pyqir",
                err_type=NotImplementedError,
            )

        context = self._llvm_module.context
        qir_function = self._external_gates_map[op_name]
        if qir_function is None:
            # First time seeing this external gate -> define new function
            qir_function_arguments = [pyqir.Type.double(context)] * len(operation.arguments)
            qir_function_arguments += [pyqir.qubit_type(context)] * op_qubit_count

            qir_function = pyqir.Function(
                pyqir.FunctionType(pyqir.Type.void(context), qir_function_arguments),
                pyqir.Linkage.EXTERNAL,
                f"__quantum__qis__{op_name}__body",
                self._llvm_module,
            )
            self._external_gates_map[op_name] = qir_function

        op_parameters = None
        if len(operation.arguments) > 0:  # parametric gate
            op_parameters = self._get_op_parameters(operation)

        if op_parameters is not None:
            self._builder.call(qir_function, [*op_parameters, *op_qubits])
        else:
            self._builder.call(qir_function, op_qubits)

    def _visit_generic_gate_operation(self, operation: qasm3_ast.QuantumGate) -> None:
        """Visit a gate operation element.

        Args:
            operation (qasm3_ast.QuantumGate): The gate operation to visit.

        Returns:
            None
        """
        if operation.name.name in self._external_gates_map:
            self._visit_external_gate_operation(operation)
        else:
            self._visit_basic_gate_operation(operation)

    def _get_branch_params(self, condition: Any) -> tuple[str, int, bool]:
        """
        Get the branch parameters from the branching condition

        Args:
            condition (Any): The condition to analyze

        Returns:
            tuple[str, int, bool]: (register name, register id, positive branch)
        """

        def validate_index_expression(expression):
            assert isinstance(expression, qasm3_ast.IndexExpression)
            assert isinstance(expression.collection, qasm3_ast.Identifier)
            assert isinstance(expression.index, list) and len(expression.index) == 1
            assert isinstance(expression.index[0], qasm3_ast.IntegerLiteral)

        if isinstance(condition, qasm3_ast.UnaryExpression):
            validate_index_expression(condition.expression)
            return (
                condition.expression.collection.name,  # type: ignore
                condition.expression.index[0].value,  # type: ignore
                not condition.op == UnaryOperator["!"],
            )
        if isinstance(condition, qasm3_ast.BinaryExpression):
            assert isinstance(
                condition.rhs, qasm3_ast.BooleanLiteral
            ), "Invalid branching condition"
            validate_index_expression(condition.lhs)
            return (
                condition.lhs.collection.name,  # type: ignore
                condition.lhs.index[0].value,  # type: ignore
                condition.rhs.value,
            )
        if isinstance(condition, qasm3_ast.IndexExpression):
            assert isinstance(condition.index, list) and len(condition.index) == 1
            return (condition.collection.name, condition.index[0].value, True)  # type: ignore
        # default case
        return "", -1, True

    def _visit_branching_statement(self, statement: qasm3_ast.BranchingStatement) -> None:
        """Visit a branching statement element.

        Args:
            statement (qasm3_ast.BranchingStatement): The branching statement to visit.

        Returns:
            None
        """

        condition = statement.condition

        if_block = statement.if_block
        else_block = statement.else_block
        reg_name, reg_id, positive_branch = self._get_branch_params(condition)

        if not positive_branch:
            if_block, else_block = else_block, if_block

        def _visit_statement_block(block):
            for stmt in block:
                self.visit_statement(stmt)

        pyqir._native.if_result(
            self._builder,
            pyqir.result(self._llvm_module.context, self._clbit_labels[f"{reg_name}_{reg_id}"]),
            zero=lambda: _visit_statement_block(else_block),
            one=lambda: _visit_statement_block(if_block),
        )

    def visit_statement(self, statement: qasm3_ast.Statement) -> None:
        """Visit a statement element.

        Args:
            statement (qasm3_ast.Statement): The statement to visit.

        Returns:
            None
        """
        logger.debug("Visiting statement '%s'", str(statement))

        visit_map = {
            qasm3_ast.Include: lambda x: None,  # No operation
            qasm3_ast.QubitDeclaration: self._visit_register,
            qasm3_ast.ClassicalDeclaration: self._visit_register,
            qasm3_ast.QuantumMeasurementStatement: self._visit_measurement,
            qasm3_ast.QuantumReset: self._visit_reset,
            qasm3_ast.QuantumBarrier: self._visit_barrier,
            qasm3_ast.QuantumGate: self._visit_generic_gate_operation,
            qasm3_ast.BranchingStatement: self._visit_branching_statement,
            qasm3_ast.QuantumPhase: lambda x: None,  # No operation
        }

        visitor_function = visit_map.get(type(statement))

        if not isinstance(statement, qasm3_ast.QuantumBarrier):
            self._check_and_apply_barrier()

        if visitor_function:
            visitor_function(statement)  # type: ignore[operator]
        else:
            raise_qasm3_error(
                f"Unsupported statement of type {type(statement)}", span=statement.span
            )

    def ir(self) -> str:
        return str(self._llvm_module)

    def bitcode(self) -> bytes:
        return self._llvm_module.bitcode
